/*
** Module   : SaxReaderImpl
** Abstract : SAX reader specific features implementation.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130208 Created initial version. Implementation is limited to the stubs to get the
**                  converted java code compiled.
** 002 CA  20130221 Added ADM-DATA and UNIQUE-ID support. 
** 003 EVL 20130320 Spelling fixes for comments.
** 004 EVK 20131001 Implement methods and class logic.
*/

package com.goldencode.p2j.xml;

import com.ctc.wstx.evt.DefaultEventAllocator;
import com.goldencode.p2j.util.*;
import org.codehaus.stax2.XMLInputFactory2;

import javax.xml.stream.*;
import javax.xml.stream.events.*;
import javax.xml.stream.util.XMLEventAllocator;
import java.io.*;
import java.util.concurrent.atomic.AtomicLong;

/**
 * Implementation SAX-Reader specific features of the SAX objects.
 */
public class SaxReaderImpl
extends SaxEntityImpl
implements SaxReader
{
   /** Current status of SAX-Writer */
   private ReadStatus readStatus = ReadStatus.SAX_UNINITIALIZED;

   /** Current input source of SAX-Reader */
   private Source source = null;

   /** Current input source of SAX-Reader */
   private InputStream inputStream = null;

   /** Sax-parser object */
   private XMLStreamReader saxParser;

   /** helps XMLStreamReader to allocate xml events*/
   private XMLEventAllocator allocator;

   /** Special wrapper to call callback procedures*/
   private XMLCallbackWrapper callbackWrapper;

   /** Specify the location of the SAX callback procedures */
   private handle hCallback;

   /** validation flag */
   private boolean validate       = true;

   /** validation flag */
   private boolean suppress       = false;

   /** Schema paths locations variable */
   private String  schemaPaths    = null;

   /** Schema locations variable */
   private String  schemaLocation = "";

   /**
    * Default constructor.
    */
   public SaxReaderImpl()
   {
   }

   /**
    * Get the value of the ADM-DATA attribute.
    *
    * @return See above.
    */
   @Override
   public character getADMData()
   {
      // TODO Auto-generated method stub
      return null;
   }

   /**
    * Set the value of the ADM-DATA attribute.
    *
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(character value)
   {
      // TODO Auto-generated method stub
   }

   /**
    * Gets the value of the handler atribute for the given SAX Reader object.
    *
    * @return   The handle variable representing the handler attribute.
    */
   @Override
   public handle getHandler()
   {
      return new handle(hCallback);
   }

   /**
    * Sets the value of the handler atribute for the given SAX Reader object.
    *
    * @param    hCallback
    *           The handle of the new callback handler to be set for the handler attribute.
    */
   @Override
   public void setHandler(handle hCallback)
   {
      if (hCallback.isUnknown())
      {
         invalidArgumentAssignError("HANDLE");
         return;
      }
      this.hCallback = new handle(hCallback);
   }
   
   /**
    * Gets the value of the current 1-based column in XML source.
    *
    * @return   The integer value of the current column.
    */
   @Override
   public integer getLocatorColumnNumber()
   {
      if (readStatus == ReadStatus.SAX_UNINITIALIZED)
      {
         return new integer();
      }
      return new integer(saxParser.getLocation().getColumnNumber());
   }

   /**
    * Gets the value of the current 1-based line number in XML source.
    *
    * @return   The integer value of the current line.
    */
   @Override
   public integer getLocatorLineNumber()
   {
      if (readStatus == ReadStatus.SAX_UNINITIALIZED)
      {
         return new integer();
      }
      return new integer(saxParser.getLocation().getLineNumber());
   }

   /**
    * Gets the current value of the public identifier of the current XML source.
    *
    * @return   The character value representing the current locator-public-id attribute.
    */
   @Override
   public character getLocatorPublicId()
   {
      if (readStatus == ReadStatus.SAX_UNINITIALIZED)
      {
         return new character();
      }
      return new character(saxParser.getLocation().getPublicId());
   }
   
   /**
    * Gets the current value of the system identifier of the current XML source.
    *
    * @return   The character value representing the current locator-system-id attribute.
    */
   @Override
   public character getLocatorSystemId()
   {
      if (readStatus == ReadStatus.SAX_UNINITIALIZED)
      {
         return new character();
      }
      return new character(saxParser.getLocation().getSystemId());
   }
   
   /**
    * Gets the current value of the parse status. The default value is
    * {@link ReadStatus#SAX_UNINITIALIZED}. The other returned values are
    * {@link ReadStatus#SAX_COMPLETE}, {@link ReadStatus#SAX_PARSER_ERROR} or
    * {@link ReadStatus#SAX_RUNNING}.
    *
    * @return   The integer value representing the current status of the parse.
    */
   @Override
   public integer getParseStatus()
   {
      return new integer(readStatus.getStatus());
   }
   
   /**
    * Gets the current value of the namespace location pair mapping.
    *
    * @return   The character value representing the current schema-location attribute.
    */
   @Override
   public character getSchemaLocation()
   {
      return new character(schemaLocation);
   }
   
   /**
    * Sets the current value of the namespace location pair mapping.
    *
    * @param    location
    *           The new value the current schema-location attribute.
    */
   @Override
   public void setSchemaLocation(String location)
   {
      setSchemaLocation(new character(location));
   }
   
   /**
    * Gets the delimiter separated list of directory paths used to locate XML DTD associated with
    * a particular XML document.
    *
    * @return   The character value representing the current schema-path attribute.
    */
   @Override
   public character getSchemaPath()
   {
      return new character(schemaPaths);
   }
   
   /**
    * Sets the delimiter separated list of directory paths used to locate XML DTD associated with
    * a particular XML document.
    * 
    * @param    path
    *           The new value the current schema-path attribute.
    */
   @Override
   public void setSchemaPath(String path)
   {
      setSchemaPath(new character(path));
   }
   
   /**
    * Indicates if the namespace processing is suppressed or not.
    * 
    * @return   The logical value representing the current suppress-namespace-processing
    *           attribute.
    */
   @Override
   public logical getSuppressNamespaceProcessing()
   {
      return new logical(suppress);
   }
   
   /**
    * Redefines whether the namespace processing is suppressed or not.
    *
    * @param    suppress
    *           The new value the current suppress-namespace-processing attribute.
    */
   @Override
   public void setSuppressNamespaceProcessing(boolean suppress)
   {
      setSuppressNamespaceProcessing(new logical(suppress));
   }
   
   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the
    * same as the handle value itself.
    *
    * @return   The integer number of the children.
    */
   @Override
   public integer getUniqueID()
   {
      // TODO Auto-generated method stub
      return null;
   }
   
   /**
    * Indicates if the parser validates the XML document against DTD or not.
    * 
    * @return   The logical value representing the current validation-enabled attribute.
    */
   @Override
   public logical getValidationEnabled()
   {
      return new logical(validate);
   }
   
   /**
    * Redefines whether the parser will validate the XML document against DTD or not.
    *
    * @param    enable
    *           The new value the current validation-enabled attribute.
    */
   @Override
   public void setValidationEnabled(boolean enable)
   {
      setValidationEnabled(new logical(enable));
   }
   
   /**
    * Executing single call parse of an XML document associated with a SAX-Reader object.
    */
   @Override
   public void saxParse()
   {
      saxParseFirst();
      while(readStatus == ReadStatus.SAX_RUNNING)
      {
         saxParseNext();
      }
   }
   
   /**
    * Initializes and begins progressive-scan parse of an XML document associated with a
    * SAX-Reader object.
    */
   @Override
   public void saxParseFirst()
   {
      if (inputStream == null)
      {
         saxParseError("SAX-PARSE-FIRST", "No input file specified");
         return;
      }

      try
      {
         inputStream.reset();
      }
      catch (IOException ignored)
      {
         saxParseError("SAX-PARSE-FIRST", "No input file specified");
         return;
      }

      try
      {
         XMLInputFactory xmlif = XMLInputFactory2.newInstance();
         xmlif.setProperty(XMLInputFactory.IS_COALESCING, Boolean.TRUE);
         xmlif.setProperty(XMLInputFactory.IS_NAMESPACE_AWARE, !suppress);
         xmlif.setProperty(XMLInputFactory.SUPPORT_DTD, validate);
         xmlif.setProperty(XMLInputFactory.IS_VALIDATING, validate);
         callbackWrapper = new XMLCallbackWrapper(hCallback, new handle(this));
         xmlif.setProperty(XMLInputFactory.REPORTER, new XMLReporter()
         {
            @Override
            public void report(String   message,
                               String   errorType,
                               Object   relatedInformation,
                               Location location)
            throws XMLStreamException
            {
               callbackWrapper.error(message);
            }
         });

         xmlif.setEventAllocator(DefaultEventAllocator.getDefaultInstance());
         saxParser = xmlif.createXMLStreamReader(inputStream);
         allocator = xmlif.getEventAllocator();

         int eventType = saxParser.getEventType();
         if (eventType != XMLStreamConstants.START_DOCUMENT)
         {
            saxParseError("SAX-PARSE-FIRST", "Unexpected error");
            return;
         }
         readStatus = ReadStatus.SAX_RUNNING;
      }
      catch (IllegalArgumentException | XMLStreamException ignored)
      {
         readStatus = ReadStatus.SAX_UNINITIALIZED;
         saxParseError("SAX-PARSE-FIRST", "Unexpected error");
      }
   }

   /**
    * Continues progressive-scan parse of an XML document associated with a SAX-Reader object.
    */
   @Override
   public void saxParseNext()
   {
      if (readStatus == ReadStatus.SAX_COMPLETE)
      {
         return;
      }

      if (readStatus == ReadStatus.SAX_UNINITIALIZED)
      {
         saxParseError("SAX-PARSE-NEXT", "No input file specified");
         return;
      }
      try
      {
         if (!saxParser.hasNext() && readStatus == ReadStatus.SAX_RUNNING)
         {
            readStatus = ReadStatus.SAX_COMPLETE;
            return;
         }
         int eventType = saxParser.next();

         XMLEvent event = getXMLEvent(saxParser);
         switch (eventType)
         {
            case XMLStreamConstants.START_ELEMENT:
               //emulate startPrefixMapping
               for (int i = 0; i < saxParser.getNamespaceCount(); i++)
               {
                  callbackWrapper.startPrefixMapping(saxParser.getNamespacePrefix(i),
                                                     saxParser.getNamespaceURI(i));
               }
               callbackWrapper.startElement(event.asStartElement());
               break;
            case XMLStreamConstants.END_ELEMENT:
               callbackWrapper.endElement(event.asEndElement());
               //emulate startPrefixMapping
               for (int i = saxParser.getNamespaceCount() - 1; i >= 0; i--)
               {
                  callbackWrapper.startPrefixMapping(saxParser.getNamespacePrefix(i),
                                                     saxParser.getNamespaceURI(i));
               }
               break;
            case XMLStreamConstants.PROCESSING_INSTRUCTION:
               callbackWrapper.processingInstruction((ProcessingInstruction)event);
               break;
            case XMLStreamConstants.CHARACTERS:
               callbackWrapper.characters(event.asCharacters());
               break;
            case XMLStreamConstants.COMMENT:
               break;
            case XMLStreamConstants.SPACE:
               callbackWrapper.ignorableWhitespace(event.asCharacters());
               break;
            case XMLStreamConstants.START_DOCUMENT:
               callbackWrapper.startDocument();
               break;
            case XMLStreamConstants.END_DOCUMENT:
               callbackWrapper.endDocument();
               break;
            case XMLStreamConstants.NOTATION_DECLARATION:
               callbackWrapper.notationDecl((NotationDeclaration)event);
               break;
            case XMLStreamConstants.DTD:
            case XMLStreamConstants.ENTITY_DECLARATION:
            case XMLStreamConstants.ENTITY_REFERENCE:
            case XMLStreamConstants.ATTRIBUTE:
            case XMLStreamConstants.CDATA:
            case XMLStreamConstants.NAMESPACE:
               break;
            default:
               saxParseError("SAX-PARSE-NEXT", "Unexpected error");
               break;
         }
      }
      catch (XMLCallbackWrapper.MismatchedParametersException e)
      {
         if (e.errorCode == ERR_10512)
         {
            mismatchCallbackNumberParameters(e.callbackName);
         }
         else if (e.errorCode == ERR_10511)
         {
            mismatchCallbackParameters("SAX-PARSE-NEXT", e.callbackName);
         }
         readStatus = ReadStatus.SAX_PARSER_ERROR;
      }
      catch (XMLStreamException ignored)
      {
         saxParseError("SAX-PARSE-NEXT", "Unexpected error");
         readStatus = ReadStatus.SAX_PARSER_ERROR;
      }
   }

   /**
    * Set the value of the ADM-DATA attribute.
    *
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(String value)
   {
      // TODO Auto-generated method stub
   }

   /**
    * Specifies the input source of the XML document to be parsed by a SAX-Reader object.
    * 
    * @param    mode
    *           The type of the object to process. Can be 'FILE', 'HANDLE', 'MEMPTR' or
    *           'LONGCHAR'.
    * @param    sourceObject
    *           The appropriate object to process.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical setInputSource(String mode, Object sourceObject)
   {
      return setInputSource(new character(mode), sourceObject);
   }

   /**
    * Specifies the input source of the XML document to be parsed by a SAX-Reader object.
    * 
    * @param    mode
    *           The type of the object to process. Can be 'FILE', 'HANDLE', 'MEMPTR' or
    *           'LONGCHAR'.
    * @param    source
    *           The appropriate object to process.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical setInputSource(character mode, Object source)
   {
      final String readMode;
      if (mode.isUnknown())
      {
         readMode = "file";
      }
      else
      {
         readMode = mode.getValue();
      }

      if ("file".equalsIgnoreCase(readMode))
      {
         this.source = Source.FILE;
         if (!(source instanceof String))
         {
            invalidArgumentError("SET-INPUT-SOURCE");
            return new logical(false);
         }
         inputStream = new FileInputStreamWrapper((String) source);
      }
      else if ("stream".equalsIgnoreCase(readMode))
      {
         this.source = Source.STREAM;
      }
      else if ("memptr".equalsIgnoreCase(readMode))
      {
         this.source = Source.MEMPTR;
         if (!(source instanceof memptr))
         {
            invalidArgumentError("SET-INPUT-SOURCE");
            return new logical(false);
         }
         inputStream = new MemptrInputStream((memptr) source);
      }
      else if ("stream-handle".equalsIgnoreCase(readMode))
      {
         this.source = Source.STREAM_HANDLE;
      }
      else if ("longchar".equalsIgnoreCase(readMode))
      {
         this.source = Source.LONGCHAR;
         if (!(source instanceof longchar))
         {
            invalidArgumentError("SET-INPUT-SOURCE");
            return new logical(false);
         }
         inputStream = new LongcharInputStream((longchar) source);
      }
      else
      {
         String errMsg = "Argument for SET-INPUT-SOURCE must be a 'file', 'longchar', 'memptr', " +
                         "or 'handle'";
         ErrorManager.recordOrShowError(ERR_10516, errMsg, false, false, false);
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Sets the current value of the namespace location pair mapping.
    *
    * @param    location
    *           The new value the current schema-location attribute.
    */
   @Override
   public void setSchemaLocation(character location)
   {
      if (location.isUnknown())
      {
         invalidArgumentError("SCHEMA-LOCATION");
      }
      schemaLocation = location.toStringMessage();
   }

   /**
    * Sets the delimiter separated list of directory paths used to locate XML DTD associated with
    * a particular XML document.
    *
    * @param    path
    *           The new value the current schema-path attribute.
    */
   @Override
   public void setSchemaPath(character path)
   {
      if (path.isUnknown())
      {
         invalidArgumentAssignError("SCHEMA-PATH");
         return;
      }

      schemaPaths = path.toStringMessage();
   }

   /**
    * Redefines whether the namespace processing is suppressed or not.
    *
    * @param    suppress
    *           The new value the current suppress-namespace-processing attribute.
    */
   @Override
   public void setSuppressNamespaceProcessing(logical suppress)
   {
      this.suppress = suppress.booleanValue();
   }

   /**
    * Redefines whether the parser will validate the XML document against DTD or not.
    *
    * @param    enable
    *           The new value the current validation-enabled attribute.
    */
   @Override
   public void setValidationEnabled(logical enable)
   {
      validate = enable.booleanValue();
   }

   /**
    * Causes parser to stop parsing the XML document.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical stopParsing()
   {
      readStatus = ReadStatus.SAX_COMPLETE;
      return new logical(true);
   }

   /**
    * Special wrapper to wrap read from longchar in lazy mode. It's means that first try to read
    * will check if longchar is available.
    */
   static class LongcharInputStream
   extends InputStream
   {
      /** A valid longchar object to read*/
      private longchar lchar;
      /** Lazy initialization flag.*/
      private volatile boolean isStarted;
      /** Buffered input Stream.*/
      private ByteArrayInputStream bais;

      /**
       * Constructor accepting a longchar, which will read after first call.
       *
       * @param    source
       *           A valid longchar object.
       */
      LongcharInputStream(longchar source)
      {
         lchar = source;
         isStarted = false;
      }

      @Override
      public int read()
      throws IOException
      {
         if (!isStarted)
         {
            isStarted = true;
            bais = new ByteArrayInputStream(lchar.toStringMessage().getBytes());
         }
         return bais.read();
      }

      /**
       * Resets the input stream to begin reading from the first character
       * of this input stream's underlying buffer.
       */
      @Override
      public synchronized void reset()
      throws IOException
      {
         isStarted = false;
      }

      @Override
      public void close()
      throws IOException
      {
         if (bais != null)
         {
            bais.close();
         }
      }
   }

   /**
    * Special wrapper to wrap read from memptr in lazy mode. It's means that first try to read
    * will check if memptr is available.
    */
   private static class MemptrInputStream
   extends InputStream
   {
      /** A valid memptr */
      private final memptr mptr;
      /** Current read cursor position, started from index 1.*/
      private final AtomicLong position = new AtomicLong();

      /**
       * Constructor accepting a memptr, which will read after first call.
       *
       * @param    source
       *           A valid memptr object.
       */
      private MemptrInputStream(memptr source)
      {
         mptr = source;
         position.set(1);
      }

      @Override
      public int read()
      throws IOException
      {
         try
         {
            return mptr.getByte(position.getAndIncrement()).intValue();
         }
         catch (ErrorConditionException e)
         {
            throw new IOException(e);
         }
      }

      /**
       * Resets the input stream to begin reading from the first character
       * of this input stream's underlying buffer.
       */
      @Override
      public synchronized void reset()
      throws IOException
      {
         position.set(1);
      }
   }

   /**
    * Special wrapper to wrap read from file in lazy mode. It's means that first try to read
    * will check if file is available.
    */
   static class FileInputStreamWrapper
   extends InputStream
   {
      /** File name.*/
      private final String fileName;
      /** Lazy initialization flag.*/
      private volatile boolean isStarted;
      /** Buffered input Stream.*/
      private Stream stream;

      /**
       * Constructor accepting a file name, which will read after first call.
       *
       * @param    source
       *           A file name.
       */
      FileInputStreamWrapper(String source)
      {
         isStarted = false;
         fileName = source;
      }

      @Override
      public int read()
      throws IOException
      {
         if (!isStarted)
         {
            try
            {
               stream = StreamFactory.openFileStream(fileName, false, false);
               if (stream == null)
               {
                  throw new IOException("An exception occurred!");
               }
            }
            catch (ErrorConditionException e)
            {
               throw new IOException("An exception occurred!", e);
            }
            isStarted = true;
         }
         int ch = stream.readCh();
         if (ch == -2)
         {
            return -1;
         }
         if (ch == -1)
         {
            throw new IOException("An exception occurred!");
         }
         return ch;
      }

      /**
       * Resets the input stream to begin reading from the first character
       * of this input stream's underlying buffer.
       */
      @Override
      public synchronized void reset()
      throws IOException
      {
         isStarted = false;
      }

      @Override
      public void close()
      throws IOException
      {
         if (stream != null)
         {
            stream.close();
         }
      }
   }

   /**
    * Process and display mismatched parameters in callback.
    *
    * @param    method
    *           The read-only attribute.
    * @param    message
    *           Additional error message.
    */
   private static void saxParseError(String method, String message)
   {
      if (method == null)
      {
         throw new IllegalArgumentException("The method is null !");
      }
      String msg = "SAX parser error: %s, %s";
      String err = String.format(msg, method.toUpperCase(), message);
      ErrorManager.recordOrThrowError(ERR_14586, err, false);
   }

   /**
    * This method allocates an event given the current state of the XMLStreamReader
    *
    * @param    reader
    *           The XMLStreamReader to allocate from
    * @return   The event corresponding to the current reader state.
    */
   private XMLEvent getXMLEvent(XMLStreamReader reader)
   throws XMLStreamException
   {
      return allocator.allocate(reader);
   }

   /**
    * Process and display mismatched parameters in callback.
    *
    * @param    method
    *           The read-only attribute.
    */
   private void mismatchCallbackParameters(String method, String callbackName)
   {
      if (method == null)
      {
         throw new IllegalArgumentException("The method is null !");
      }
      String msg = "Callback error: %s sent procedure %s mismatched parameters";
      String err = String.format(msg, method.toUpperCase(), callbackName);
      ErrorManager.recordOrThrowError(ERR_10512, err, false);
   }

   /**
    * Process and display mismatched parameters in callback.
    *
    * @param    callbackName
    *           name of callback.
    */
   private void mismatchCallbackNumberParameters(String callbackName)
   {
      if (callbackName == null)
      {
         throw new IllegalArgumentException("The callbackName is null !");
      }
      String msg = "Callback error: mismatched number of parameters passed to procedure %s.";
      String err = String.format(msg, callbackName);
      ErrorManager.recordOrThrowError(ERR_10511, err, false);
   }
}